# Copyright (c) 2023 Darren Erik Vengroff
"""Places in Alabama."""


ABANDA_CDP = "00100"
ABBEVILLE_CITY = "00124"
ADAMSVILLE_CITY = "00460"
ADDISON_TOWN = "00484"
AKRON_TOWN = "00676"
ALABASTER_CITY = "00820"
ALBERTVILLE_CITY = "00988"
ALEXANDER_CITY_CITY = "01132"
ALEXANDRIA_CDP = "01180"
ALICEVILLE_CITY = "01228"
ALLGOOD_TOWN = "01396"
ALTOONA_TOWN = "01660"
ANDALUSIA_CITY = "01708"
ANDERSON_TOWN = "01756"
ANNISTON_CITY = "01852"
ARAB_CITY = "02116"
ARDMORE_TOWN = "02260"
ARGO_TOWN = "02320"
ARITON_TOWN = "02428"
ARLEY_TOWN = "02500"
ASHFORD_TOWN = "02836"
ASHLAND_TOWN = "02860"
ASHVILLE_CITY = "02908"
ATHENS_CITY = "02956"
ATMORE_CITY = "03004"
ATTALLA_CITY = "03028"
AUBURN_CITY = "03076"
AUTAUGAVILLE_TOWN = "03220"
AVON_TOWN = "03364"
AXIS_CDP = "03508"
BABBIE_TOWN = "03556"
BAILEYTON_TOWN = "03676"
BAKERHILL_TOWN = "03724"
BALLPLAY_CDP = "03820"
BANKS_TOWN = "03940"
BAY_MINETTE_CITY = "04660"
BAYOU_LA_BATRE_CITY = "04684"
BEAR_CREEK_TOWN = "04852"
BEATRICE_TOWN = "04900"
BEAVERTON_TOWN = "04948"
BELGREEN_CDP = "05140"
BELK_TOWN = "05164"
BELLAMY_CDP = "05188"
BELLE_FONTAINE_CDP = "05212"
BENTON_TOWN = "05692"
BERLIN_TOWN = "05812"
BERRY_TOWN = "05932"
BESSEMER_CITY = "05980"
BILLINGSLEY_TOWN = "06460"
BIRMINGHAM_CITY = "07000"
BLACK_TOWN = "07120"
BLOUNTSVILLE_TOWN = "07456"
BLUE_RIDGE_CDP = "07588"
BLUE_SPRINGS_TOWN = "07672"
BOAZ_CITY = "07912"
BOLIGEE_TOWN = "08104"
BON_AIR_TOWN = "08248"
BON_SECOUR_CDP = "08272"
BOYKIN_CDP = "08680"
BRANTLEY_TOWN = "09016"
BRANTLEYVILLE_CDP = "09064"
BRENT_CITY = "09136"
BREWTON_CITY = "09208"
BRIDGEPORT_CITY = "09328"
BRIGHTON_CITY = "09400"
BRILLIANT_TOWN = "09424"
BRISTOW_COVE_CDP = "09457"
BROOK_HIGHLAND_CDP = "09500"
BROOKSIDE_TOWN = "09736"
BROOKWOOD_TOWN = "09808"
BROOMTOWN_CDP = "09880"
BRUNDIDGE_CITY = "10240"
BUCKS_CDP = "10456"
BUTLER_TOWN = "11032"
CALERA_CITY = "11416"
CALVERT_CDP = "11488"
CAMDEN_CITY = "11512"
CAMP_HILL_TOWN = "11680"
CARBON_HILL_CITY = "12016"
CARDIFF_TOWN = "12040"
CARLISLE_ROCKLEDGE_CDP = "12070"
CARLTON_CDP = "12112"
CAROLINA_TOWN = "12160"
CARROLLTON_TOWN = "12304"
CASTLEBERRY_TOWN = "12496"
CATHERINE_CDP = "12544"
CEDAR_BLUFF_TOWN = "12760"
CENTER_POINT_CITY = "13264"
CENTRE_CITY = "13648"
CENTREVILLE_CITY = "13672"
CHATOM_TOWN = "14008"
CHELSEA_CITY = "14104"
CHEROKEE_TOWN = "14152"
CHICKASAW_CITY = "14392"
CHILDERSBURG_CITY = "14464"
CHOCCOLOCCO_CDP = "14728"
CHUNCHULA_CDP = "14968"
CITRONELLE_CITY = "15064"
CLANTON_CITY = "15136"
CLAY_CITY = "15256"
CLAYHATCHEE_TOWN = "15304"
CLAYTON_TOWN = "15376"
CLEVELAND_TOWN = "15472"
CLIO_CITY = "15640"
COALING_TOWN = "16000"
COATS_BEND_CDP = "16080"
COFFEE_SPRINGS_TOWN = "16240"
COFFEEVILLE_TOWN = "16264"
COKER_TOWN = "16312"
COLLINSVILLE_TOWN = "16600"
COLONY_TOWN = "16684"
COLUMBIA_TOWN = "16744"
COLUMBIANA_CITY = "16768"
CONCORD_CDP = "16936"
COOSADA_TOWN = "17176"
CORDOVA_CITY = "17368"
COTTONDALE_CDP = "17704"
COTTONWOOD_TOWN = "17824"
COUNTY_LINE_TOWN = "17968"
COURTLAND_TOWN = "18040"
COWARTS_TOWN = "18088"
CREOLA_CITY = "18304"
CROSSVILLE_TOWN = "18856"
CUBA_TOWN = "18952"
CULLMAN_CITY = "18976"
CULLOMBURG_CDP = "19024"
CUSSETA_TOWN = "19216"
DADEVILLE_CITY = "19336"
DALEVILLE_CITY = "19360"
DAPHNE_CITY = "19648"
DAUPHIN_ISLAND_TOWN = "19744"
DAVISTON_TOWN = "19816"
DAYTON_TOWN = "19912"
DEATSVILLE_TOWN = "20008"
DECATUR_CITY = "20104"
DEER_PARK_CDP = "20152"
DELTA_CDP = "20272"
DEMOPOLIS_CITY = "20296"
DETROIT_TOWN = "20392"
DODGE_CITY_TOWN = "20955"
DORA_CITY = "21136"
DOTHAN_CITY = "21184"
DOUBLE_SPRINGS_TOWN = "21280"
DOUGLAS_TOWN = "21328"
DOZIER_TOWN = "21448"
DUNNAVANT_CDP = "21890"
DUTTON_TOWN = "21952"
EAGLE_POINT_CDP = "22088"
EAST_BREWTON_CITY = "22216"
EAST_POINT_CDP = "22504"
ECLECTIC_TOWN = "22816"
EDGEWATER_CDP = "23032"
EDWARDSVILLE_TOWN = "23176"
EGYPT_CDP = "23224"
ELBA_CITY = "23296"
ELBERTA_TOWN = "23320"
ELDRIDGE_TOWN = "23344"
ELKMONT_TOWN = "23488"
ELMORE_TOWN = "23656"
EMELLE_TOWN = "23872"
EMERALD_MOUNTAIN_CDP = "23880"
ENTERPRISE_CITY = "24184"
EPES_TOWN = "24256"
EQUALITY_CDP = "24280"
ETHELSVILLE_TOWN = "24472"
EUFAULA_CITY = "24568"
EUNOLA_CDP = "24616"
EUTAW_TOWN = "24664"
EVA_TOWN = "24688"
EVERGREEN_CITY = "24808"
EXCEL_TOWN = "24880"
FAIRFIELD_CITY = "25120"
FAIRFORD_CDP = "25216"
FAIRHOPE_CITY = "25240"
FAIRVIEW_TOWN = "25384"
FALKVILLE_TOWN = "25648"
FAUNSDALE_TOWN = "25816"
FAYETTE_CITY = "25840"
FAYETTEVILLE_CDP = "25864"
FITZPATRICK_CDP = "26152"
FIVE_POINTS_TOWN = "26200"
FLOMATON_TOWN = "26824"
FLORALA_TOWN = "26848"
FLORENCE_CITY = "26896"
FOLEY_CITY = "26992"
FORESTDALE_CDP = "27088"
FORKLAND_TOWN = "27376"
FORT_DEPOSIT_TOWN = "27520"
FORT_PAYNE_CITY = "27616"
FORT_RUCKER_CDP = "27640"
FRANKLIN_TOWN = "28024"
FREDONIA_CDP = "28120"
FRISCO_CITY_TOWN = "28312"
FRUITDALE_CDP = "28408"
FRUITHURST_TOWN = "28432"
FULTON_TOWN = "28504"
FULTONDALE_CITY = "28552"
FYFFE_TOWN = "28672"
GADSDEN_CITY = "28696"
GAINESVILLE_TOWN = "28768"
GALLANT_CDP = "28816"
GANTT_TOWN = "28936"
GARDEN_CITY_TOWN = "29032"
GARDENDALE_CITY = "29056"
GAYLESVILLE_TOWN = "29296"
GEIGER_TOWN = "29392"
GENEVA_CITY = "29464"
GEORGIANA_TOWN = "29560"
GERALDINE_TOWN = "29608"
GILBERTOWN_TOWN = "29704"
GLEN_ALLEN_TOWN = "29944"
GLENCOE_CITY = "29992"
GLENWOOD_TOWN = "30160"
GOLDVILLE_TOWN = "30448"
GOOD_HOPE_CITY = "30496"
GOODWATER_TOWN = "30640"
GORDO_TOWN = "30736"
GORDON_TOWN = "30760"
GORDONVILLE_TOWN = "30808"
GOSHEN_TOWN = "30880"
GRAHAM_CDP = "31000"
GRAND_BAY_CDP = "31024"
GRANT_TOWN = "31096"
GRAYSON_VALLEY_CDP = "31348"
GRAYSVILLE_CITY = "31384"
GREENSBORO_CITY = "31720"
GREENVILLE_CITY = "31912"
GRIMES_TOWN = "32056"
GROVE_HILL_TOWN = "32080"
GU_WIN_TOWN = "32536"
GUIN_CITY = "32224"
GULF_SHORES_CITY = "32272"
GULFCREST_CDP = "32248"
GUNTERSVILLE_CITY = "32416"
GURLEY_TOWN = "32440"
HACKLEBURG_TOWN = "32560"
HACKNEYVILLE_CDP = "32584"
HALEBURG_TOWN = "32656"
HALEYVILLE_CITY = "32704"
HAMILTON_CITY = "32848"
HAMMONDVILLE_TOWN = "32896"
HANCEVILLE_CITY = "32968"
HANOVER_CDP = "33064"
HARPERSVILLE_TOWN = "33256"
HARTFORD_CITY = "33424"
HARTSELLE_CITY = "33448"
HARVEST_CDP = "33472"
HATTON_CDP = "33544"
HAYDEN_TOWN = "33640"
HAYNEVILLE_TOWN = "33712"
HAZEL_GREEN_CDP = "33808"
HEADLAND_CITY = "33856"
HEATH_TOWN = "33904"
HEFLIN_CITY = "33976"
HELENA_CITY = "34024"
HENAGAR_CITY = "34096"
HIGHLAND_LAKE_TOWN = "34480"
HIGHLAND_LAKES_CDP = "34490"
HILLSBORO_TOWN = "34816"
HISSOP_CDP = "34984"
HOBSON_CDP = "35130"
HOBSON_CITY_TOWN = "35152"
HODGES_TOWN = "35200"
HOKES_BLUFF_CITY = "35392"
HOLLINS_CDP = "35512"
HOLLIS_CROSSROADS_CDP = "35536"
HOLLY_POND_TOWN = "35560"
HOLLYWOOD_TOWN = "35632"
HOLT_CDP = "35704"
HOLTVILLE_CDP = "35752"
HOMEWOOD_CITY = "35800"
HOOVER_CITY = "35896"
HORN_HILL_TOWN = "36088"
HUEYTOWN_CITY = "36448"
HUGULEY_CDP = "36592"
HUNTSVILLE_CITY = "37000"
HURTSBORO_TOWN = "37096"
HYTOP_TOWN = "37264"
IDER_TOWN = "37312"
INDIAN_SPRINGS_VILLAGE_CITY = "37465"
IRONDALE_CITY = "37864"
IVALEE_CDP = "38032"
JACKSON_CITY = "38152"
JACKSONS_GAP_TOWN = "38248"
JACKSONVILLE_CITY = "38272"
JASPER_CITY = "38416"
JEMISON_CITY = "38608"
JOPPA_CDP = "39064"
KANSAS_TOWN = "39280"
KELLYTON_TOWN = "39472"
KENNEDY_TOWN = "39520"
KILLEN_TOWN = "39784"
KIMBERLY_CITY = "39856"
KINSEY_TOWN = "40072"
KINSTON_TOWN = "40096"
LA_FAYETTE_CITY = "40672"
LADONIA_CDP = "40648"
LAKE_VIEW_TOWN = "40928"
LAKEVIEW_TOWN = "40888"
LANETT_CITY = "41296"
LANGSTON_TOWN = "41368"
LEEDS_CITY = "41968"
LEESBURG_TOWN = "42016"
LEIGHTON_TOWN = "42160"
LEROY_CDP = "42304"
LESTER_TOWN = "42352"
LEVEL_PLAINS_TOWN = "42472"
LEXINGTON_TOWN = "42640"
LIBERTYVILLE_TOWN = "42808"
LILLIAN_CDP = "42928"
LINCOLN_CITY = "43120"
LINDEN_CITY = "43240"
LINEVILLE_CITY = "43264"
LIPSCOMB_CITY = "43336"
LISMAN_TOWN = "43384"
LITTLEVILLE_TOWN = "43648"
LIVINGSTON_CITY = "43720"
LOACHAPOKA_TOWN = "43744"
LOCKHART_TOWN = "43816"
LOCUST_FORK_TOWN = "43888"
LOOKOUT_MOUNTAIN_CDP = "44160"
LOUISVILLE_TOWN = "44344"
LOWNDESBORO_TOWN = "44536"
LOXLEY_TOWN = "44608"
LUVERNE_CITY = "44728"
LYNN_TOWN = "44800"
MACEDONIA_CDP = "45316"
MADISON_CITY = "45784"
MADRID_TOWN = "45904"
MAGNOLIA_SPRINGS_TOWN = "46072"
MALCOLM_CDP = "46192"
MALVERN_TOWN = "46264"
MAPLESVILLE_TOWN = "46504"
MARBURY_CDP = "46600"
MARGARET_TOWN = "46696"
MARION_CITY = "46768"
MAYTOWN_TOWN = "47728"
MCCALLA_CDP = "45040"
MCDONALD_CHAPEL_CDP = "45208"
MCINTOSH_TOWN = "45472"
MCKENZIE_TOWN = "45496"
MCMULLEN_TOWN = "45640"
MEADOWBROOK_CDP = "47740"
MEGARGEL_CDP = "47848"
MEMPHIS_TOWN = "48052"
MENTONE_TOWN = "48064"
MERIDIANVILLE_CDP = "48112"
MIDFIELD_CITY = "48376"
MIDLAND_CITY_TOWN = "48400"
MIDWAY_TOWN = "48424"
MIGNON_CDP = "48616"
MILLBROOK_CITY = "48712"
MILLERVILLE_CDP = "48784"
MILLPORT_TOWN = "48808"
MILLRY_TOWN = "48832"
MINOR_CDP = "49072"
MOBILE_CITY = "50000"
MONROEVILLE_CITY = "50192"
MONTEVALLO_CITY = "50312"
MONTGOMERY_CITY = "51000"
MOODY_CITY = "51096"
MOORES_MILL_CDP = "51216"
MOORESVILLE_TOWN = "51264"
MORRIS_TOWN = "51456"
MORRISON_CROSSROADS_CDP = "51469"
MOSSES_TOWN = "51520"
MOULTON_CITY = "51600"
MOUNDVILLE_TOWN = "51648"
MOUNT_OLIVE_CDP_COOSA_COUNTY = "52320"
MOUNT_OLIVE_CDP_JEFFERSON_COUNTY = "52344"
MOUNT_VERNON_TOWN = "52608"
MOUNTAIN_BROOK_CITY = "51696"
MOVICO_CDP = "52680"
MULGA_TOWN = "52776"
MUNFORD_TOWN = "52848"
MUSCLE_SHOALS_CITY = "53016"
MYRTLEWOOD_TOWN = "53112"
NANAFALIA_CDP = "53208"
NANCES_CREEK_CDP = "53232"
NAPIER_FIELD_TOWN = "53256"
NATURAL_BRIDGE_TOWN = "53376"
NAUVOO_TOWN = "53400"
NECTAR_TOWN = "53448"
NEEDHAM_TOWN = "53472"
NEW_BROCKTON_TOWN = "53856"
NEW_HOPE_CITY = "54168"
NEW_MARKET_CDP = "54312"
NEW_SITE_TOWN = "54432"
NEW_UNION_CDP = "54582"
NEWBERN_TOWN = "53784"
NEWTON_TOWN = "54480"
NEWVILLE_TOWN = "54600"
NIXBURG_CDP = "54696"
NORTH_COURTLAND_TOWN = "55044"
NORTH_JOHNS_TOWN = "55152"
NORTHPORT_CITY = "55200"
NOTASULGA_TOWN = "55440"
OAK_GROVE_TOWN = "55752"
OAK_HILL_TOWN = "55848"
OAKMAN_TOWN = "55992"
ODENVILLE_CITY = "56400"
OHATCHEE_TOWN = "56472"
ONEONTA_CITY = "57000"
ONYCHA_TOWN = "57024"
OPELIKA_CITY = "57048"
OPP_CITY = "57120"
ORANGE_BEACH_CITY = "57144"
ORRVILLE_TOWN = "57240"
OUR_TOWN_CDP = "57384"
OWENS_CROSS_ROADS_TOWN = "57504"
OXFORD_CITY = "57576"
OZARK_CITY = "57648"
PAINT_ROCK_TOWN = "57696"
PANOLA_CDP = "57912"
PARRISH_TOWN = "58272"
PEA_RIDGE_CDP = "58680"
PELHAM_CITY = "58848"
PELL_CITY_CITY = "58896"
PENNINGTON_TOWN = "58968"
PENTON_CDP = "59016"
PERDIDO_BEACH_TOWN = "59088"
PERDIDO_CDP = "59064"
PETERMAN_CDP = "59232"
PETREY_TOWN = "59328"
PHENIX_CITY_CITY = "59472"
PHIL_CAMPBELL_TOWN = "59496"
PICKENSVILLE_TOWN = "59592"
PIEDMONT_CITY = "59640"
PIKE_ROAD_TOWN = "59760"
PINCKARD_TOWN = "59832"
PINE_APPLE_TOWN = "59880"
PINE_HILL_TOWN = "60240"
PINE_LEVEL_CDP = "60264"
PINE_RIDGE_TOWN = "60372"
PINSON_CITY = "60648"
PISGAH_TOWN = "60720"
PLEASANT_GROVE_CITY = "61008"
PLEASANT_GROVES_TOWN = "61038"
POINT_CLEAR_CDP = "61488"
POLLARD_TOWN = "61536"
POWELL_TOWN = "62075"
PRATTVILLE_CITY = "62328"
PRICEVILLE_TOWN = "62472"
PRICHARD_CITY = "62496"
PROVIDENCE_TOWN = "62688"
PUTNAM_CDP = "62952"
RAGLAND_TOWN = "63216"
RAINBOW_CITY_CITY = "63288"
RAINSVILLE_CITY = "63336"
RANBURNE_TOWN = "63408"
RAY_CDP = "63495"
RED_BAY_CITY = "63576"
RED_LEVEL_TOWN = "63768"
REDLAND_CDP = "63730"
REDSTONE_ARSENAL_CDP = "63912"
REECE_CITY_TOWN = "63984"
REELTOWN_CDP = "64040"
REFORM_TOWN = "64104"
REHOBETH_TOWN = "64152"
REMLAP_CDP = "64248"
REPTON_TOWN = "64368"
RIDGEVILLE_TOWN = "64656"
RIVER_FALLS_TOWN = "64776"
RIVERSIDE_CITY = "64920"
RIVERVIEW_TOWN = "65016"
ROANOKE_CITY = "65040"
ROBERTSDALE_CITY = "65208"
ROCK_CREEK_CDP = "65412"
ROCK_MILLS_CDP = "65592"
ROCKFORD_TOWN = "65472"
ROCKVILLE_CDP = "65784"
ROGERSVILLE_TOWN = "66216"
ROSA_TOWN = "66408"
RUSSELLVILLE_CITY = "67056"
RUTLEDGE_TOWN = "67176"
SAKS_CDP = "67608"
SAMSON_CITY = "67800"
SAND_ROCK_TOWN = "67920"
SANFORD_TOWN = "68016"
SARALAND_CITY = "68160"
SARDIS_CITY_TOWN = "68280"
SATSUMA_CITY = "68352"
SCOTTSBORO_CITY = "68736"
SECTION_TOWN = "69000"
SELMA_CITY = "69120"
SELMONT_WEST_SELMONT_CDP = "69180"
SEMMES_CITY = "69240"
SHEFFIELD_CITY = "69648"
SHELBY_CDP = "69672"
SHILOH_TOWN = "69840"
SHOAL_CREEK_CDP = "69978"
SHORTER_TOWN = "70128"
SILAS_TOWN = "70416"
SILVERHILL_TOWN = "70536"
SIMS_CHAPEL_CDP = "70632"
SIPSEY_TOWN = "70704"
SKYLINE_TOWN = "70896"
SLOCOMB_CITY = "71040"
SMITHS_STATION_CITY = "71190"
SMOKE_RISE_CDP = "71201"
SNEAD_TOWN = "71280"
SOMERVILLE_TOWN = "71496"
SOUTH_VINEMONT_TOWN = "71900"
SOUTHSIDE_CITY = "71832"
SPANISH_FORT_CITY = "71976"
SPRING_GARDEN_CDP = "72408"
SPRINGVILLE_CITY = "72600"
SPRUCE_PINE_CDP = "72672"
ST_FLORIAN_TOWN = "67536"
ST_STEPHENS_CDP = "67584"
STANDING_ROCK_CDP = "72720"
STAPLETON_CDP = "72816"
STEELE_TOWN = "72888"
STERRETT_CDP = "73056"
STEVENSON_CITY = "73080"
STEWARTVILLE_CDP = "73152"
STOCKTON_CDP = "73272"
SULLIGENT_CITY = "73728"
SUMITON_CITY = "73848"
SUMMERDALE_TOWN = "73872"
SUSAN_MOORE_TOWN = "74160"
SWEET_WATER_TOWN = "74304"
SYLACAUGA_CITY = "74352"
SYLVAN_SPRINGS_TOWN = "74424"
SYLVANIA_TOWN = "74400"
TALLADEGA_SPRINGS_TOWN = "74616"
TALLADEGA_CITY = "74592"
TALLASSEE_CITY = "74688"
TARRANT_CITY = "74976"
TAYLOR_TOWN = "75096"
THEODORE_CDP = "75768"
THOMASTON_TOWN = "75936"
THOMASVILLE_CITY = "75960"
THORSBY_TOWN = "76080"
TIBBIE_CDP = "76200"
TIDMORE_BEND_CDP = "76208"
TILLMANS_CORNER_CDP = "76320"
TOWN_CREEK_TOWN = "76584"
TOXEY_TOWN = "76632"
TRAFFORD_TOWN = "76680"
TRIANA_TOWN = "76824"
TRINITY_TOWN = "76872"
TROY_CITY = "76920"
TRUSSVILLE_CITY = "76944"
TUSCALOOSA_CITY = "77256"
TUSCUMBIA_CITY = "77280"
TUSKEGEE_CITY = "77304"
TWIN_TOWN = "77352"
UNDERWOOD_PETERSVILLE_CDP = "77580"
UNION_GROVE_TOWN = "77784"
UNION_SPRINGS_CITY = "77880"
UNION_TOWN = "77616"
UNIONTOWN_TOWN = "77904"
URIAH_CDP = "78120"
VALLEY_GRANDE_CITY = "78232"
VALLEY_HEAD_TOWN = "78240"
VALLEY_CITY = "78204"
VANCE_TOWN = "78264"
VANDIVER_CDP = "78312"
VERNON_CITY = "78480"
VESTAVIA_HILLS_CITY = "78552"
VINA_TOWN = "78984"
VINCENT_TOWN = "79008"
VINEGAR_BEND_CDP = "79032"
VREDENBURGH_TOWN = "79272"
WADLEY_TOWN = "79344"
WALDO_TOWN = "79488"
WALNUT_GROVE_TOWN = "79728"
WARRIOR_CITY = "79944"
WATERLOO_TOWN = "80064"
WAVERLY_TOWN = "80256"
WEAVER_CITY = "80352"
WEBB_TOWN = "80376"
WEDOWEE_TOWN = "80496"
WEOGUFKA_CDP = "80736"
WEST_BLOCTON_TOWN = "80928"
WEST_END_COBB_TOWN_CDP = "81084"
WEST_JEFFERSON_TOWN = "81336"
WEST_POINT_TOWN = "81528"
WESTOVER_TOWN = "81504"
WETUMPKA_CITY = "81720"
WHATLEY_CDP = "81744"
WHITE_HALL_TOWN = "81912"
WHITE_PLAINS_CDP = "82104"
WHITESBORO_CDP = "82176"
WILSONVILLE_TOWN = "82848"
WILTON_TOWN = "82872"
WINFIELD_CITY = "82992"
WOODLAND_TOWN = "83400"
WOODSTOCK_TOWN = "83640"
WOODVILLE_TOWN = "83664"
YELLOW_BLUFF_TOWN = "83976"
YORK_CITY = "84096"



# Copyright (c) 2023 Darren Erik Vengroff
"""Places in New Mexico."""


ABEYTAS_CDP = "00240"
ABIQUIU_CDP = "00310"
ACOMITA_LAKE_CDP = "00765"
ADELINO_CDP = "00940"
AGUA_FRIA_CDP = "01220"
ALAMILLO_CDP = "01640"
ALAMO_CDP = "01710"
ALAMOGORDO_CITY = "01780"
ALBUQUERQUE_CITY = "02000"
ALCALDE_CDP = "02070"
ALGODONES_CDP = "02210"
ALMA_CDP = "02560"
ANGEL_FIRE_VILLAGE = "03400"
ANGUSTURA_CDP = "03680"
ANIMAS_CDP = "03750"
ANTHONY_CITY = "03820"
ANTON_CHICO_CDP = "03890"
ANZAC_VILLAGE_CDP = "03970"
APACHE_CREEK_CDP = "04030"
ARAGON_CDP = "04450"
ARENAS_VALLEY_CDP = "04660"
ARREY_CDP = "04870"
ARROYO_HONDO_CDP_SANTA_FE_COUNTY = "05002"
ARROYO_HONDO_CDP_TAOS_COUNTY = "05010"
ARROYO_SECO_CDP = "05150"
ARTESIA_CITY = "05220"
ATOKA_CDP = "05430"
AZTEC_CITY = "05780"
BARTON_CDP = "06220"
BAYARD_CITY = "06270"
BECENTI_CDP = "06340"
BECLABITO_CDP = "06425"
BELEN_CITY = "06480"
BENT_CDP = "06760"
BERINO_CDP = "06830"
BERNALILLO_TOWN = "06970"
BIBO_CDP = "07145"
BLACK_HAT_CDP = "07492"
BLACK_ROCK_CDP = "07670"
BLANCO_CDP = "07810"
BLOOMFIELD_CITY = "07880"
BLUEWATER_CDP = "08180"
BLUEWATER_VILLAGE_CDP = "08190"
BOLES_ACRES_CDP = "08440"
BORREGO_PASS_CDP = "08493"
BOSQUE_FARMS_VILLAGE = "08580"
BRAZOS_CDP = "08860"
BRIMHALL_NIZHONI_CDP = "09035"
BROADVIEW_CDP = "09202"
BUCKHORN_CDP = "09630"
BUTTERFIELD_PARK_CDP = "09980"
CABALLO_CDP = "10050"
CANDY_KITCHEN_CDP = "10592"
CANJILON_CDP = "10610"
CANNON_AFB_CDP = "10750"
CANOVA_CDP = "11520"
CAPITAN_VILLAGE = "11800"
CAPULIN_CDP = "12010"
CARLSBAD_CITY = "12150"
CARNUEL_CDP = "12430"
CARRIZOZO_TOWN = "12500"
CASA_COLORADA_CDP = "12786"
CASAS_ADOBES_CDP = "12833"
CATALPA_CANYON_CDP = "12844"
CAUSEY_VILLAGE = "12850"
CANADA_DE_LOS_ALAMOS_CDP = "10470"
CANON_CDP = "10770"
CANONCITO_CDP = "11256"
CANONES_CDP = "11310"
CEDAR_CREST_CDP = "13200"
CEDAR_GROVE_CDP = "13270"
CEDAR_HILL_CDP = "13340"
CEDRO_CDP = "13450"
CENTER_POINT_CDP = "13540"
CHAMA_VILLAGE = "13970"
CHAMBERINO_CDP = "14040"
CHAMISAL_CDP = "14110"
CHAMITA_CDP = "14180"
CHAMIZAL_CDP = "14200"
CHAPARRAL_CDP = "14250"
CHICAL_CDP = "14670"
CHILI_CDP = "14810"
CHILILI_CDP = "14880"
CHIMAYO_CDP = "14950"
CHUPADERO_CDP = "15300"
CHURCH_ROCK_CDP = "15370"
CIMARRON_VILLAGE = "15440"
CITY_OF_THE_SUN_CDP = "15560"
CLAYTON_TOWN = "15720"
CLIFF_CDP = "15860"
CLOUDCROFT_VILLAGE = "16280"
CLOVIS_CITY = "16420"
COBRE_CDP = "16490"
COCHITI_CDP = "16560"
COCHITI_LAKE_CDP = "16585"
COLUMBUS_VILLAGE = "17050"
CONCHAS_DAM_CDP = "17260"
CONEJO_CDP = "17295"
CONTINENTAL_DIVIDE_CDP = "17330"
CORDOVA_CDP = "17610"
CORONA_VILLAGE = "17680"
CORRALES_VILLAGE = "17960"
COSTILLA_CDP = "18240"
COTTON_CITY_CDP = "18310"
COYOTE_CDP = "18660"
CRESTVIEW_CDP = "18770"
CROUCH_MESA_CDP = "18905"
CROWNPOINT_CDP = "18940"
CRUZVILLE_CDP = "19010"
CRYSTAL_CDP = "19080"
CUARTELEZ_CDP = "19110"
CUBA_VILLAGE = "19150"
CUBERO_CDP = "19220"
CUNDIYO_CDP = "19500"
CUYAMUNGUE_CDP = "19640"
CUYAMUNGUE_GRANT_CDP = "19645"
DATIL_CDP = "19920"
DEER_CANYON_CDP = "20015"
DEMING_CITY = "20270"
DES_MOINES_VILLAGE = "20480"
DEXTER_TOWN = "20620"
DIXON_CDP = "20830"
DORA_VILLAGE = "21180"
DONA_ANA_CDP = "21110"
DULCE_CDP = "21390"
DURAN_CDP = "21810"
EAGLE_NEST_VILLAGE = "22020"
EAST_PECOS_CDP = "22230"
EDGEWOOD_TOWN = "22380"
EDITH_ENCLAVE_CDP = "22389"
EL_CERRO_CDP = "22580"
EL_CERRO_MISSION_CDP = "22582"
EL_DUENDE_CDP = "22650"
EL_MORRO_VALLEY_CDP = "23497"
EL_RANCHO_CDP = "23840"
EL_RITO_CDP = "24120"
EL_VALLE_DE_ARROYO_SECO_CDP = "24347"
ELDORADO_AT_SANTA_FE_CDP = "22625"
ELEPHANT_BUTTE_CITY = "22720"
ELIDA_TOWN = "23000"
ENCANTADO_CDP = "24453"
ENCINAL_CDP = "24540"
ENCINO_VILLAGE = "24610"
ENSENADA_CDP = "24960"
ESCONDIDA_CDP = "25100"
ESCUDILLA_BONITA_CDP = "25135"
ESPANOLA_CITY = "25170"
ESTANCIA_TOWN = "25380"
EUNICE_CITY = "25450"
FAIRACRES_CDP = "25590"
FARMINGTON_CITY = "25800"
FAYWOOD_CDP = "25870"
FENCE_LAKE_CDP = "26010"
FLORA_VISTA_CDP = "26360"
FLOYD_VILLAGE = "26570"
FOLSOM_VILLAGE = "26780"
FORT_SUMNER_VILLAGE = "27340"
FORT_WINGATE_CDP = "27410"
FRUITLAND_CDP = "27760"
GALISTEO_CDP = "27970"
GALLINA_CDP = "28180"
GALLUP_CITY = "28460"
GAMERCO_CDP = "28530"
GARFIELD_CDP = "28670"
GILA_CDP = "29160"
GILA_HOT_SPRINGS_CDP = "29230"
GLEN_ACRES_CDP = "29540"
GLENWOOD_CDP = "29790"
GLORIETA_CDP = "29860"
GOLDEN_ACRES_CDP = "30035"
GOLDEN_CDP = "30000"
GRADY_VILLAGE = "30210"
GRANTS_CITY = "30490"
GRENVILLE_VILLAGE = "30770"
HACHITA_CDP = "31190"
HAGERMAN_TOWN = "31330"
HANOVER_CDP = "31540"
HAPPY_VALLEY_CDP = "31680"
HATCH_VILLAGE = "31820"
HAYSTACK_CDP = "31925"
HERNANDEZ_CDP = "32100"
HIGH_ROLLS_CDP = "32160"
HIGHLAND_MEADOWS_CDP = "32148"
HILLSBORO_CDP = "32380"
HOBBS_CITY = "32520"
HOLLOMAN_AFB_CDP = "32940"
HOMER_C_JONES_CDP = "33122"
HOMESTEAD_CDP = "33130"
HOPE_VILLAGE = "33290"
HOT_SPRINGS_LANDING_CDP = "33640"
HOUSE_VILLAGE = "33710"
HURLEY_TOWN = "33850"
HYDE_PARK_CDP = "33920"
INDIAN_HILLS_CDP = "34430"
ISLETA_CDP = "34620"
IYANBITO_CDP = "34830"
JACONA_CDP = "34900"
JACONITA_CDP = "34925"
JAL_CITY = "35040"
JAMESTOWN_CDP = "35070"
JARALES_CDP = "35110"
JEMEZ_PUEBLO_CDP = "35250"
JEMEZ_SPRINGS_VILLAGE = "35320"
KEELER_FARM_CDP = "35595"
KINGSTON_CDP = "36090"
KIRTLAND_AFB_CDP = "36300"
KIRTLAND_TOWN = "36230"
LA_BAJADA_CDP = "36555"
LA_BOCA_CDP = "36567"
LA_CIENEGA_CDP = "36720"
LA_CUEVA_CDP_SANDOVAL_COUNTY = "36880"
LA_CUEVA_CDP_SANTA_FE_COUNTY = "36930"
LA_HACIENDA_CDP = "37407"
LA_HUERTA_CDP = "37420"
LA_JARA_CDP = "37490"
LA_JOYA_CDP = "37700"
LA_LUZ_CDP = "38330"
LA_MADERA_CDP_RIO_ARRIBA_COUNTY = "38470"
LA_MADERA_CDP_SANDOVAL_COUNTY = "38540"
LA_MESA_CDP = "38680"
LA_MESILLA_CDP = "38820"
LA_PLATA_CDP = "39030"
LA_PUEBLA_CDP = "39170"
LA_TIERRA_CDP = "40074"
LA_UNION_CDP = "40220"
LA_VILLITA_CDP = "40360"
LAGUNA_CDP = "37070"
LAKE_ARTHUR_TOWN = "37840"
LAKE_ROBERTS_CDP = "37860"
LAKE_ROBERTS_HEIGHTS_CDP = "37862"
LAKE_SUMNER_CDP = "37875"
LAKE_VALLEY_CDP = "37910"
LAMY_CDP = "38890"
LAS_CAMPANAS_CDP = "39332"
LAS_CRUCES_CITY = "39380"
LAS_MARAVILLAS_CDP = "39430"
LAS_NUTRIAS_CDP = "39450"
LAS_PALOMAS_CDP = "39520"
LAS_TUSAS_CDP = "39792"
LAS_VEGAS_CITY = "39940"
LEE_ACRES_CDP = "40525"
LEMITAR_CDP = "40570"
LINDRITH_CDP = "41130"
LITTLE_WALNUT_VILLAGE_CDP = "41340"
LIVINGSTON_WHEELER_CDP = "41420"
LLANO_DEL_MEDIO_CDP = "41620"
LOBO_CANYON_CDP = "41953"
LOCO_HILLS_CDP = "41970"
LOGAN_VILLAGE = "42040"
LORDSBURG_CITY = "42180"
LOS_ALAMOS_CDP = "42320"
LOS_CERRILLOS_CDP = "42600"
LOS_CHAVES_CDP = "42740"
LOS_LUCEROS_CDP = "43300"
LOS_LUNAS_VILLAGE = "43370"
LOS_OJOS_CDP = "43510"
LOS_RANCHOS_DE_ALBUQUERQUE_VILLAGE = "43930"
LOVING_VILLAGE = "44420"
LOVINGTON_CITY = "44490"
LOWER_FRISCO_CDP = "44595"
LUIS_LOPEZ_CDP = "45190"
LUMBERTON_CDP = "45260"
LUNA_CDP = "45330"
LYBROOK_CDP = "45365"
LYDEN_CDP = "45400"
MADRID_CDP = "46100"
MADRONE_CDP = "46170"
MAGDALENA_VILLAGE = "46310"
MALAGA_CDP = "46380"
MANUELITO_CDP = "46660"
MANZANO_CDP = "46730"
MANZANO_SPRINGS_CDP = "46770"
MAXWELL_VILLAGE = "47080"
MAYHILL_CDP = "47150"
MCCARTYS_VILLAGE_CDP = "45690"
MCGAFFEY_CDP = "45890"
MCINTOSH_CDP = "46030"
MEADOW_LAKE_CDP = "47220"
MEDANALES_CDP = "47360"
MELROSE_VILLAGE = "47570"
MESCALERO_CDP = "47920"
MESILLA_TOWN = "48060"
MESITA_CDP = "48200"
MESQUITE_CDP = "48270"
MIDDLE_FRISCO_CDP = "48425"
MIDDLE_MESA_CDP = "48435"
MIDWAY_CDP = "48480"
MILAN_VILLAGE = "48620"
MIMBRES_CDP = "48830"
MOGOLLON_CDP = "49180"
MONTEREY_PARK_CDP = "49515"
MONUMENT_CDP = "49950"
MOQUINO_CDP = "50020"
MORA_CDP = "50090"
MORIARTY_CITY = "50160"
MORNINGSIDE_CDP = "50195"
MOSQUERO_VILLAGE = "50300"
MOUNT_TAYLOR_CDP = "50615"
MOUNTAIN_VIEW_CDP_CIBOLA_COUNTY = "50512"
MOUNTAIN_VIEW_CDP_LUNA_COUNTY = "50515"
MOUNTAINAIR_TOWN = "50370"
NADINE_CDP = "50720"
NAGEEZI_CDP = "50790"
NAKAIBITO_CDP = "50860"
NAMBE_CDP = "50930"
NAPI_HEADQUARTERS_CDP = "51035"
NARA_VISA_CDP = "51140"
NASCHITTI_CDP = "51280"
NAVAJO_CDP = "51420"
NAVAJO_DAM_CDP = "51490"
NENAHNEZAD_CDP = "51715"
NEWCOMB_CDP = "51770"
NEWKIRK_CDP = "51840"
NOGAL_CDP = "52190"
NORTH_ACOMITA_VILLAGE_CDP = "52300"
NORTH_HOBBS_CDP = "52500"
NORTH_HURLEY_CDP = "52540"
NORTH_LIGHT_PLANT_CDP = "52560"
NORTH_SAN_YSIDRO_CDP = "52680"
NORTH_VALLEY_CDP = "52750"
OASIS_CDP = "53010"
OHKAY_OWINGEH_CDP = "53150"
OJO_AMARILLO_CDP = "53485"
OJO_CALIENTE_CDP = "53590"
OJO_ENCINO_CDP = "53620"
OJO_SARCO_CDP = "53730"
OLD_TOWN_CDP = "54097"
ORGAN_CDP = "54430"
OROGRANDE_CDP = "54500"
PAA_KO_CDP = "54675"
PAGUATE_CDP = "54710"
PAJARITO_MESA_CDP = "54940"
PARADISE_HILLS_CDP = "55270"
PARAJE_CDP = "55340"
PASTURA_CDP = "55550"
PEAK_PLACE_CDP = "55595"
PECAN_PARK_CDP = "55605"
PECOS_VILLAGE = "55620"
PERALTA_TOWN = "56180"
PENA_BLANCA_CDP = "55830"
PENASCO_CDP = "55900"
PICACHO_HILLS_CDP = "56748"
PICURIS_PUEBLO_CDP = "56810"
PIE_TOWN_CDP = "56950"
PINEDALE_CDP = "57160"
PINEHAVEN_CDP = "57175"
PINEHILL_CDP = "57190"
PINON_CDP = "57440"
PINOS_ALTOS_CDP = "57510"
PLACITAS_CDP_DONA_ANA_COUNTY = "57990"
PLACITAS_CDP_SANDOVAL_COUNTY = "58070"
PLAYAS_CDP = "58210"
PLEASANTON_CDP = "58420"
POJOAQUE_CDP = "58630"
POLVADERA_CDP = "58910"
PONDEROSA_CDP = "58980"
PONDEROSA_PINE_CDP = "59120"
PORTALES_CITY = "59260"
PREWITT_CDP = "59540"
PUEBLITO_CDP = "59750"
PUEBLITOS_CDP = "59820"
PUEBLO_CDP = "59840"
PUEBLO_PINTADO_CDP = "60000"
PUEBLO_OF_SANDIA_VILLAGE_CDP = "59950"
PUERTO_DE_LUNA_CDP = "60170"
PULPOTIO_BAREAS_CDP = "60205"
PUNTA_DE_AGUA_CDP = "60310"
PURTY_ROCK_CDP = "60327"
QUEMADO_CDP = "60730"
QUESTA_VILLAGE = "60870"
RADIUM_SPRINGS_CDP = "61010"
RAMAH_CDP = "61220"
RANCHO_GRANDE_CDP = "61675"
RANCHOS_DE_TAOS_CDP = "61710"
RATON_CITY = "62060"
RED_RIVER_TOWN = "62200"
RED_ROCK_RANCH_CDP = "62295"
REDROCK_CDP = "62270"
REGINA_CDP = "62340"
RESERVE_VILLAGE = "62620"
RIBERA_CDP = "62760"
RINCON_CDP = "62900"
RIO_CHIQUITO_CDP = "63110"
RIO_COMMUNITIES_CITY = "63145"
RIO_LUCIO_CDP = "63250"
RIO_RANCHO_ESTATES_CDP = "63530"
RIO_RANCHO_CITY = "63460"
RIO_EN_MEDIO_CDP = "63168"
RIVERS_CDP = "63660"
ROCK_SPRINGS_CDP = "64230"
RODEO_CDP = "64370"
RODEY_CDP = "64440"
ROSEDALE_CDP = "64900"
ROSWELL_CITY = "64930"
ROWE_CDP = "65000"
ROY_VILLAGE = "65070"
RUIDOSO_DOWNS_CITY = "65280"
RUIDOSO_VILLAGE = "65210"
SACRAMENTO_CDP = "65560"
SAGAR_CDP = "65595"
SALEM_CDP = "65840"
SAN_ACACIA_CDP = "65910"
SAN_ANTONIO_CDP = "66120"
SAN_ANTONITO_CDP_BERNALILLO_COUNTY = "66330"
SAN_ANTONITO_CDP_SOCORRO_COUNTY = "66400"
SAN_CRISTOBAL_CDP = "66540"
SAN_FELIPE_PUEBLO_CDP = "67450"
SAN_FIDEL_CDP = "67520"
SAN_ILDEFONSO_PUEBLO_CDP = "68010"
SAN_JON_VILLAGE = "68150"
SAN_JOSE_CDP_RIO_ARRIBA_COUNTY = "68430"
SAN_JOSE_CDP_SAN_MIGUEL_COUNTY = "68500"
SAN_LORENZO_CDP = "68920"
SAN_LUIS_CDP = "69060"
SAN_MATEO_CDP = "69270"
SAN_MIGUEL_CDP = "69410"
SAN_PABLO_CDP = "69645"
SAN_PEDRO_CDP = "69825"
SAN_RAFAEL_CDP = "69875"
SAN_YSIDRO_CDP = "71015"
SAN_YSIDRO_VILLAGE = "71020"
SANDIA_HEIGHTS_CDP = "66765"
SANDIA_KNOLLS_CDP = "66890"
SANDIA_PARK_CDP = "66960"
SANOSTEE_CDP = "69620"
SANTA_ANA_PUEBLO_CDP = "70250"
SANTA_CLARA_PUEBLO_CDP = "70390"
SANTA_CLARA_VILLAGE = "70270"
SANTA_CRUZ_CDP = "70460"
SANTA_FE_FOOTHILLS_CDP = "70505"
SANTA_FE_CITY = "70500"
SANTA_ROSA_CITY = "70670"
SANTA_TERESA_CDP = "70700"
SANTO_DOMINGO_PUEBLO_CDP = "70810"
SAUSAL_CDP = "71145"
SEAMA_CDP = "71440"
SEBOYETA_CDP = "71510"
SEDILLO_CDP = "71700"
SENA_CDP = "71790"
SETON_VILLAGE_CDP = "72140"
SHEEP_SPRINGS_CDP = "72600"
SHIPROCK_CDP = "72770"
SILVER_CITY_TOWN = "73260"
SKYLINE_GANIPA_CDP = "73440"
SOCORRO_CITY = "73540"
SOHAM_CDP = "73680"
SOMBRILLO_CDP = "73890"
SOUTH_ACOMITA_VILLAGE_CDP = "73925"
SOUTH_RIVER_CDP = "74275"
SOUTH_VALLEY_CDP = "74520"
SPENCERVILLE_CDP = "74660"
SPRINGER_TOWN = "74800"
STANLEY_CDP = "75010"
STONERIDGE_CDP = "75465"
SUNDANCE_CDP = "75605"
SUNLAND_PARK_CITY = "75640"
SUNLIT_HILLS_CDP = "75650"
SUNSHINE_CDP = "75710"
TAJIQUE_CDP = "76060"
TALPA_CDP = "76130"
TANO_ROAD_CDP = "76165"
TAOS_PUEBLO_CDP = "76410"
TAOS_SKI_VALLEY_VILLAGE = "76480"
TAOS_TOWN = "76200"
TATUM_TOWN = "76620"
TECOLOTE_CDP = "76760"
TECOLOTITO_CDP = "76830"
TESUQUE_CDP = "77040"
TESUQUE_PUEBLO_CDP = "77100"
TEXICO_CITY = "77250"
THOREAU_CDP = "77530"
THUNDER_MOUNTAIN_CDP = "77635"
TIERRA_AMARILLA_CDP = "77670"
TIJERAS_VILLAGE = "77880"
TIMBERLAKE_CDP = "77930"
TIMBERON_CDP = "77950"
TOHATCHI_CDP = "78440"
TOME_CDP = "78650"
TORREON_CDP_SANDOVAL_COUNTY = "79070"
TORREON_CDP_TORRANCE_COUNTY = "79140"
TORTUGAS_CDP = "79210"
TOTAH_VISTA_CDP = "79264"
TRES_ARROYOS_CDP = "79500"
TROUT_VALLEY_CDP = "79680"
TRUCHAS_CDP = "79700"
TRUTH_OR_CONSEQUENCES_CITY = "79840"
TSE_BONITO_CDP = "79885"
TUCUMCARI_CITY = "79910"
TULAROSA_VILLAGE = "79980"
TURLEY_CDP = "80120"
TWIN_FORKS_CDP = "80470"
TWIN_LAKES_CDP = "80540"
TYRONE_CDP = "80820"
UNIVERSITY_PARK_CDP = "81030"
UPPER_FRUITLAND_CDP = "81177"
UTE_PARK_CDP = "81450"
VADITO_CDP = "81520"
VADO_CDP = "81590"
VALENCIA_CDP_SANTA_FE_COUNTY = "81775"
VALENCIA_CDP_VALENCIA_COUNTY = "81800"
VALLE_VISTA_CDP = "82332"
VANDERWAGEN_CDP = "82430"
VAUGHN_TOWN = "82570"
VEGUITA_CDP = "82640"
VELARDE_CDP = "82710"
VENTURA_CDP = "82800"
VILLANUEVA_CDP = "83130"
VIRDEN_VILLAGE = "83200"
WAGON_MOUND_VILLAGE = "83340"
WATERFLOW_CDP = "83760"
WATROUS_CDP = "83830"
WEED_CDP = "83970"
WEST_HAMMOND_CDP = "84200"
WHITE_CLIFFS_CDP = "84505"
WHITE_ROCK_CDP = "84740"
WHITE_SANDS_CDP = "84845"
WHITE_SIGNAL_CDP = "84975"
WHITES_CITY_CDP = "84950"
WILLARD_VILLAGE = "85160"
WILLIAMS_ACRES_CDP = "85230"
WILLIAMSBURG_VILLAGE = "85300"
WINDMILL_CDP = "85485"
WINSTON_CDP = "85720"
YAH_TA_HEY_CDP = "85860"
YOUNGSVILLE_CDP = "86210"
ZIA_PUEBLO_CDP = "86420"
ZUNI_PUEBLO_CDP = "86595"



# Copyright (c) 2023 Darren Erik Vengroff
"""Utilities for geometric operations."""

import math
from typing import Optional, TypeVar, Union

import geopandas as gpd
from shapely import MultiPolygon, Polygon


def isoperimetric_quotient(
    geo: Union[gpd.GeoDataFrame, gpd.GeoSeries, Polygon]
) -> float:
    """
    Compute the isoperimetric quotient of a shape or collection of shapes.

    The isoperimetric quotient is a measure of the compactness of a polygom.
    It is defines as the ratio of the area of the shape to the area of a
    circle with the same perimeter. It ranges from zero, when the polygon
    has zero area, to 1.0 when the polygon approximates a circle. It is not
    possible for a shape that is not a circle to have more area than a circle
    with the same permimeter.

    For more info, see https://en.wikipedia.org/wiki/Compactness_measure_of_a_shape
    and https://en.wikipedia.org/wiki/Isoperimetric_inequality.

    This is also equivalent to the Polsby–Popper test. See
    https://en.wikipedia.org/wiki/Polsby%E2%80%93Popper_test.

    Parameters
    ----------
    geo:
        The geography or a series or dataframe of them.

    Returns
    -------
        The isoperimetric_quotient (Polsby–Popper test) value or values
        between 0 and 1.
    """
    area = geo.area
    length = geo.exterior.length

    q = 4 * math.pi * area / (length * length)

    return q


def drop_polygon_if_sliver(
    polygon: Polygon, threshold: float = 0.01
) -> Optional[Polygon]:
    """
    Drop a polygon if it is a sliver.

    Parameters
    ----------
    polygon
        The polygon to check.
    threshold
        Threshold of sliveryness (isoperimetric quotient).

    Returns
    -------
        `polygon` if not a sliver or `None` if a sliver.
    """
    if isoperimetric_quotient(polygon) < threshold:
        return None

    return polygon


def drop_slivers_multi_polygon(
    multi_polygon: Union[Polygon, MultiPolygon], threshold: float = 0.01
) -> Optional[Union[Polygon, MultiPolygon]]:
    """
    Drop all the sliver polygons from a multi-polygon.

    Parameters
    ----------
    multi_polygon
        A `MultiPolygon`.
    threshold
        The isopermimetric threshold below which polygons are considered
        slivers.

    Returns
    -------
        A shape (either a `Polygon` or `MultiPolygon` containing all non-sliver
        polygones. If there aren't any, `None` is returned.
    """
    remaining_polygons = [
        poly
        for poly in multi_polygon.geoms
        if isoperimetric_quotient(poly) >= threshold
    ]

    if len(remaining_polygons) == 0:
        return None
    elif len(remaining_polygons) == 1:
        return remaining_polygons[0]
    else:
        return MultiPolygon(remaining_polygons)


def drop_slivers_from_geo_series(
    gs_geo: gpd.GeoSeries, threshold: float = 0.01
) -> gpd.GeoSeries:
    """
    Drop all slivers from the geometries in a `GeoSeries`.

    Parameters
    ----------
    gs_geo
        The original `GeoSeries`.
    threshold
        The isoperimetric quotient threshold.

    Returns
    -------
        The series with all slivers removed.
    """
    return gs_geo.map(
        lambda s: (
            drop_slivers_multi_polygon(s, threshold)
            if isinstance(s, MultiPolygon)
            else drop_polygon_if_sliver(s, threshold) if isinstance(s, Polygon) else s
        )
    )


def drop_slivers_from_gdf(
    gdf_geo: gpd.GeoDataFrame, threshold: float = 0.01
) -> gpd.GeoDataFrame:
    """
    Drop all slivers from the geometries in a `GeoDataFrame`.

    Parameters
    ----------
    gdf_geo
        The original `GeoDataFrame`.
    threshold
        The isoperimetric quotient threshold.

    Returns
    -------
        The geo data frame with all slivers removed.
    """
    new_geometry = drop_slivers_from_geo_series(gdf_geo.geometry, threshold=threshold)

    gdf_result = gpd.GeoDataFrame(gdf_geo, geometry=new_geometry)

    return gdf_result


T = TypeVar("T", Polygon, MultiPolygon, gpd.GeoSeries, gpd.GeoDataFrame)


def drop_slivers(
    geo: T,
    threshold: float = 0.01,
) -> Union[None, T]:
    """
    Drop slivers from an arbitrary geometry or collection of geometries.

    Accepts `MultiPolygon`, `Polygon`, `gpd.GeoSeries` or `gpd.GeoDataFrame`

    Parameters
    ----------
    geo
        the geography.
    threshold
        the isoperimetric threshold.

    Returns
    -------
        The geometry with slivers dropped.
    """
    if isinstance(geo, MultiPolygon):
        return drop_slivers_multi_polygon(geo, threshold=threshold)
    elif isinstance(geo, Polygon):
        return drop_slivers_multi_polygon(MultiPolygon((geo,)), threshold=threshold)
    if isinstance(geo, gpd.GeoSeries):
        return drop_slivers_from_geo_series(geo, threshold=threshold)
    if isinstance(geo, gpd.GeoDataFrame):
        return drop_slivers_from_gdf(geo, threshold=threshold)

    raise ValueError(
        f"Unable to drop slivers from a {type(geo)}. "
        "Must be a GeoDataFrame, GeoSeries, and MultiPolygones only."
    )



# Copyright (c) 2023 Darren Erik Vengroff
"""Counties in Idaho."""


ADA = "001"
ADAMS = "003"
BANNOCK = "005"
BEAR_LAKE = "007"
BENEWAH = "009"
BINGHAM = "011"
BLAINE = "013"
BOISE = "015"
BONNER = "017"
BONNEVILLE = "019"
BOUNDARY = "021"
BUTTE = "023"
CAMAS = "025"
CANYON = "027"
CARIBOU = "029"
CASSIA = "031"
CLARK = "033"
CLEARWATER = "035"
CUSTER = "037"
ELMORE = "039"
FRANKLIN = "041"
FREMONT = "043"
GEM = "045"
GOODING = "047"
IDAHO = "049"
JEFFERSON = "051"
JEROME = "053"
KOOTENAI = "055"
LATAH = "057"
LEMHI = "059"
LEWIS = "061"
LINCOLN = "063"
MADISON = "065"
MINIDOKA = "067"
NEZ_PERCE = "069"
ONEIDA = "071"
OWYHEE = "073"
PAYETTE = "075"
POWER = "077"
SHOSHONE = "079"
TETON = "081"
TWIN_FALLS = "083"
VALLEY = "085"
WASHINGTON = "087"



# Copyright (c) 2023 Darren Erik Vengroff
"""Counties in Michigan."""


ALCONA = "001"
ALGER = "003"
ALLEGAN = "005"
ALPENA = "007"
ANTRIM = "009"
ARENAC = "011"
BARAGA = "013"
BARRY = "015"
BAY = "017"
BENZIE = "019"
BERRIEN = "021"
BRANCH = "023"
CALHOUN = "025"
CASS = "027"
CHARLEVOIX = "029"
CHEBOYGAN = "031"
CHIPPEWA = "033"
CLARE = "035"
CLINTON = "037"
CRAWFORD = "039"
DELTA = "041"
DICKINSON = "043"
EATON = "045"
EMMET = "047"
GENESEE = "049"
GLADWIN = "051"
GOGEBIC = "053"
GRAND_TRAVERSE = "055"
GRATIOT = "057"
HILLSDALE = "059"
HOUGHTON = "061"
HURON = "063"
INGHAM = "065"
IONIA = "067"
IOSCO = "069"
IRON = "071"
ISABELLA = "073"
JACKSON = "075"
KALAMAZOO = "077"
KALKASKA = "079"
KENT = "081"
KEWEENAW = "083"
LAKE = "085"
LAPEER = "087"
LEELANAU = "089"
LENAWEE = "091"
LIVINGSTON = "093"
LUCE = "095"
MACKINAC = "097"
MACOMB = "099"
MANISTEE = "101"
MARQUETTE = "103"
MASON = "105"
MECOSTA = "107"
MENOMINEE = "109"
MIDLAND = "111"
MISSAUKEE = "113"
MONROE = "115"
MONTCALM = "117"
MONTMORENCY = "119"
MUSKEGON = "121"
NEWAYGO = "123"
OAKLAND = "125"
OCEANA = "127"
OGEMAW = "129"
ONTONAGON = "131"
OSCEOLA = "133"
OSCODA = "135"
OTSEGO = "137"
OTTAWA = "139"
PRESQUE_ISLE = "141"
ROSCOMMON = "143"
SAGINAW = "145"
SANILAC = "151"
SCHOOLCRAFT = "153"
SHIAWASSEE = "155"
ST_CLAIR = "147"
ST_JOSEPH = "149"
TUSCOLA = "157"
VAN_BUREN = "159"
WASHTENAW = "161"
WAYNE = "163"
WEXFORD = "165"



# Copyright (c) 2023 Darren Erik Vengroff
"""Counties in Maine."""


ANDROSCOGGIN = "001"
AROOSTOOK = "003"
CUMBERLAND = "005"
FRANKLIN = "007"
HANCOCK = "009"
KENNEBEC = "011"
KNOX = "013"
LINCOLN = "015"
OXFORD = "017"
PENOBSCOT = "019"
PISCATAQUIS = "021"
SAGADAHOC = "023"
SOMERSET = "025"
WALDO = "027"
WASHINGTON = "029"
YORK = "031"



# Copyright (c) 2022, 2023 Darren Erik Vengroff
"""
Utilities for loading census data.

This module relies on the US Census API, which
it wraps in a pythonic manner.
"""

import warnings
from logging import getLogger
from typing import (
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
)

import geopandas as gpd
import numpy as np
import pandas as pd

import censusdis.geography as cgeo
import censusdis.maps as cmap
from censusdis.impl.exceptions import CensusApiException
from censusdis.impl.fetch import data_from_url
from censusdis.impl.us_census_shapefiles import (
    add_geography,
    clip_water,
    infer_geo_level,
    geo_query_from_data_query_inner_geo,
)
from censusdis.impl.varcache import VariableCache
from censusdis.impl.varsource.base import VintageType
from censusdis.impl.varsource.censusapi import CensusApiVariableSource
from censusdis.values import ALL_SPECIAL_VALUES

import censusdis.impl.fetch


logger = getLogger(__name__)


GeoFilterType = Optional[Union[str, Iterable[str]]]
"""
The type we accept for geographic filters.

They are used for the values of `kwargs` to
:py:func:`download`.

These filters are either single values as a string,
or, if multivalued, then an iterable containing all
the values allowed by the filter. For example::

    import censusdis.data as ced

    from censusdis.states import NJ, NY, CT

    # Two different kinds of kwarg for `state=`, both of
    # which are of `GeoFilterType`:
    df_one_state = ced.download("aca/acs5", 2020, ["NAME"], state=NJ)
    df_tri_state = ced.download("aca/acs5", 2020, ["NAME"], state=[NJ, NY, CT])
"""


def _gf2s(geo_filter: GeoFilterType) -> Optional[str]:
    """
    Convert a filter to a string.

    For the Census API, multiple values are encoded
    in a single comma separated string.
    """
    if geo_filter is None or isinstance(geo_filter, str):
        return geo_filter
    return ",".join(geo_filter)


_MAX_VARIABLES_PER_DOWNLOAD = 50
"""
The maximum number of variables we can ask for in one census API query.

The U.S. Census sets this limit, not us. In order to not expose our
users to the limit, :py:func:`~download` mostly obscures the fact that
requests to download more than this many variables are broken into
multiple calls to the census API and then the results are stitched back
together be either merging or concatenation. This is all handled in
:py:func:`~_download_multiple`.
"""


__dw_strategy_metrics = {"merge": 0, "concat": 0}
"""
Counters for how often we use each strategy for wide tables.
"""


def _download_wide_strategy_metrics() -> Dict[str, int]:
    """
    Metrics on which strategies have been used for wide tables.

    Returns
    -------
        A dictionary of metrics on how often each strategy has
        been used.
    """
    return dict(**__dw_strategy_metrics)


def _download_multiple(
    dataset: str,
    vintage: VintageType,
    download_variables: List[str],
    *,
    key: Optional[str],
    census_variables: "VariableCache",
    with_geometry: bool = False,
    row_keys: Optional[Union[str, Iterable[str]]] = None,
    **kwargs: cgeo.InSpecType,
) -> pd.DataFrame:
    """
    Download data in groups of columns and concatenate the results together.

    The reason for this function is that the API will only return a maximum
    of 50 columns per query. This function downloads wider data 50 columns
    at a time and concatenates them.

    Parameters
    ----------
    dataset
        The dataset to download from. For example `"acs/acs5"`,
        `"dec/pl"`, or `"timeseries/poverty/saipe/schdist"`.
    vintage
        The vintage to download data for. For most data sets this is
        an integer year, for example, `2020`. But for
        a timeseries data set, pass the string `'timeseries'`.
    download_variables
        The census variables to download, for example `["NAME", "B01001_001E"]`.
    with_geometry
        If `True` a :py:class:`gpd.GeoDataFrame` will be returned and each row
        will have a geometry that is a cartographic boundary suitable for platting
        a map. See https://www.census.gov/geographies/mapping-files/time-series/geo/cartographic-boundary.2020.html
        for details of the shapefiles that will be downloaded on your behalf to
        generate these boundaries.
    api_key
        An optional API key. If you don't have or don't use a key, the number
        of calls you can make will be limited.
    variable_cache
        A cache of metadata about variables.
    row_keys
        An optional set of identifier keys to help merge together requests for more than the census API limit of
        50 variables per query. These keys are useful for census datasets such as the Current Population Survey
        where the geographic identifiers do not uniquely identify each row.
    kwargs
        A specification of the geometry that we want data for.

    Returns
    -------
        The full results of the query with all columns.

    """
    # Divide the variables into groups. If row keys are provided, include them in each chunk of variables,
    # while respecting the variable max
    if row_keys:
        chunk_size = _MAX_VARIABLES_PER_DOWNLOAD - len(row_keys)
        variable_groups = [
            # black and flake8 disagree about the whitespace before ':' here...
            # We need to drop duplicates in each chunk of variables
            # since the row_key variables might already be present in one of the chunks
            [
                item
                for item in row_keys
                + download_variables[start : start + chunk_size]  # noqa: E203
                if item not in row_keys
                or row_keys.index(item)
                == (
                    row_keys
                    + download_variables[start : start + chunk_size]  # noqa: E203
                ).index(item)
            ]
            for start in range(0, len(download_variables), chunk_size)
        ]
    else:
        variable_groups = [
            # black and flake8 disagree about the whitespace before ':' here...
            download_variables[start : start + _MAX_VARIABLES_PER_DOWNLOAD]  # noqa: 203
            for start in range(0, len(download_variables), _MAX_VARIABLES_PER_DOWNLOAD)
        ]

    if len(variable_groups) < 2:
        raise ValueError(
            "_download_multiple expects to be called with at least "
            f"{_MAX_VARIABLES_PER_DOWNLOAD + 1} variables. With fewer,"
            "use download instead."
        )

    # Get the data for each chunk.
    dfs = [
        download(
            dataset,
            vintage,
            variable_group,
            api_key=key,
            variable_cache=census_variables,
            with_geometry=with_geometry and (ii == 0),
            **kwargs,
        )
        for ii, variable_group in enumerate(variable_groups)
    ]

    # What variables came back in the first df but were not
    # requested? These are a key to the geography the row
    # represents. For example, 'STATE' amd 'COUNTY' might
    # be these variables if we did a county-level query to
    # the census API.
    geo_key_variables = [f for f in dfs[0].columns if f not in set(variable_groups[0])]

    # If we put in the geometry column, it's not part of the
    # key.
    if with_geometry:
        geo_key_variables = [f for f in geo_key_variables if f != "geometry"]

    # Now we have to decide if we are going to use the merge
    # strategy or the concat strategy to combine the data frames
    # we downloaded. Why do we have two strategies? Because we are
    # dealing with two kinds of data. One kind, from data sets like
    # ACS (https://www.census.gov/programs-surveys/acs.html),
    # has a unique key of columns that specify geography. The other
    # kind, from data sets like CPS
    # (https://www.census.gov/programs-surveys/cps.html) doesn't.
    #
    # In the unique key case, we can join the data frames that come
    # back on those key columns and get the final wide data frame
    # we want for the user.
    #
    # In the non-unique key case, we can't do this. There data sets
    # may have multiple rows for a value of the key columns. We can't
    # join here. Instead, we can only concatenate the tables
    # horizontally and hope that the rows came back in the same order
    # for each of them.

    # We hope to be able to merge. It is safer. If row_keys is supplied, they are included in
    # merge keys
    merge_strategy = True
    if row_keys:
        merge_keys = geo_key_variables + row_keys
    else:
        merge_keys = geo_key_variables
    # But if there are any non-unique keys in any df, we can't
    # merge.
    for df_slice in dfs:
        if len(df_slice.value_counts(merge_keys, sort=False)) != len(df_slice.index):
            merge_strategy = False
            break

    if merge_strategy:
        # We can do the merge strategy.

        __dw_strategy_metrics["merge"] = __dw_strategy_metrics["merge"] + 1

        df_data = dfs[0]

        for df_right in dfs[1:]:
            df_data = df_data.merge(df_right, on=merge_keys)
    else:
        # We are going to have to fall back on the concat
        # strategy. Before we do the concat, however, let's
        # double-check that the key columns are the same in
        # at the corresponding row in every df. Otherwise, something
        # is fishy, and it is not safe to concat without mixing
        # data that should be in different rows.

        rows0 = len(dfs[0].index)

        for df_slice in dfs[1:]:
            if not (
                rows0 == len(df_slice.index)
                and dfs[0][geo_key_variables].equals(df_slice[geo_key_variables])
            ):
                # At least one difference. So we cannot use the
                # concat strategy either.
                if not row_keys:
                    raise CensusApiException(
                        "Neither the merge nor the concat strategy is viable. "
                        "We made multiple queries to the census API because more than "
                        f"{_MAX_VARIABLES_PER_DOWNLOAD} variables were requested. "
                        "If you don't need all the variables, it is always safer to "
                        f"download less than {_MAX_VARIABLES_PER_DOWNLOAD} variables. "
                        f"If you need more than {_MAX_VARIABLES_PER_DOWNLOAD}, you can supply the `row_keys`"
                        "arguement with a set of variables that uniquely identify each row."
                    )
                else:
                    raise CensusApiException(
                        f"Neither the merge nor the concat strategy is viable using row_keys: {row_keys}. "
                        "The supplied keys should uniquely identify every row in the dataset to work. "
                        "If you don't need all the variables, it is always safer to "
                        f"download less than {_MAX_VARIABLES_PER_DOWNLOAD} variables. "
                    )

        # Concat strategy is as safe as it will ever be. We hope the server
        # side did not reorder the results across queries.
        logger.info(
            "Using the concat strategy, which is not guaranteed reliable if "
            "the census API returned data for multiple sub-queries of less than "
            "or equal to %d in different row orders. "
            "It is always safest to query no more than %d "
            "variables at a time. Please do so unless you really need them all.",
            _MAX_VARIABLES_PER_DOWNLOAD,
            _MAX_VARIABLES_PER_DOWNLOAD,
        )

        __dw_strategy_metrics["concat"] = __dw_strategy_metrics["concat"] + 1

        df_data = pd.concat(
            [dfs[0]] + [df.drop(geo_key_variables, axis="columns") for df in dfs[1:]],
            axis="columns",
        )

    return df_data


def download(
    dataset: str,
    vintage: VintageType,
    download_variables: Optional[Union[str, Iterable[str]]] = None,
    *,
    group: Optional[Union[str, Iterable[str]]] = None,
    leaves_of_group: Optional[Union[str, Iterable[str]]] = None,
    set_to_nan: Union[bool, Iterable[int]] = True,
    skip_annotations: bool = True,
    with_geometry: bool = False,
    remove_water: bool = False,
    download_contained_within: Optional[Dict[str, cgeo.InSpecType]] = None,
    area_threshold: float = 0.8,
    api_key: Optional[str] = None,
    variable_cache: Optional["VariableCache"] = None,
    row_keys: Optional[Union[str, Iterable[str]]] = None,
    **kwargs: cgeo.InSpecType,
) -> Union[pd.DataFrame, gpd.GeoDataFrame]:
    """
    Download data from the US Census API.

    This is the main API for downloading US Census data with the
    `censusdis` package. There are many examples of how to use
    this in the demo notebooks provided with the package at
    https://github.com/vengroff/censusdis/tree/main/notebooks.

    *A note on variables and groups*: there are multiple ways to specify the
    variables you want to download, either individually in `download_variables`,
    by one or more groups in `group`, and by the leaves of one or more groups
    in `leaves_of_group`. Note that these three sources af variables are
    deduplicated, so you will only get one column for a variable no matter
    how many times it is specified.

    *Specifying census geographies*: censusdis provides access to many
    census datasets, each of which can be retrieved at a particular set of
    geographic grains. To accomodate this, `download()` takes a set
    of kwargs to define the geographic level of the returned data. You can check
    which geographies are available for a particular dataset with the
    `geographies()`.

    Parameters
    ----------
    dataset
        The dataset to download from. For example `"acs/acs5"`,
        `"dec/pl"`, or `"timeseries/poverty/saipe/schdist"`. There are
        symbolic names for datasets, like `ACS5` for `"acs/acs5"
        in :py:module:`censusdis.datasets`.
    vintage
        The vintage to download data for. For most data sets this is
        an integer year, for example, `2020`. But for
        a timeseries data set, pass the string `'timeseries'`.
    download_variables
        The census variables to download, for example `["NAME", "B01001_001E"]`.
    group
        One or more groups (as defined by the U.S. Census for the data set)
        whose variable values should be downloaded. These are in addition to
        any specified in `download_variables`.
    leaves_of_group
        One or more groups (as defined by the U.S. Census for the data set)
        whose leaf variable values should be downloaded.These are in addition to
        any specified in `download_variables` or `group`. See
        :py:meth:`VariableCache.group_leaves` for more details on the semantics of
        leaves vs. non-leaf group variables.
    set_to_nan
        A list of values that should be set to NaN. Normally these are special
        values that the U.S. Census API sometimes returns. If `True`, then all
        values in :py:ref:`censusdis.values.ALL_SPECIAL_VALUES` will be replaced.
        If `False`, no replacements will be made.
    skip_annotations
        If `True` try to filter out `group` or `leaves_of_group` variables that are
        annotations rather than actual values. See :py:meth:`VariableCache.group_variables`
        for more details. Variable names passed in `download_variables` are not
        affected by this flag.
    with_geometry
        If `True` a :py:class:`gpd.GeoDataFrame` will be returned and each row
        will have a geometry that is a cartographic boundary suitable for platting
        a map. See https://www.census.gov/geographies/mapping-files/time-series/geo/cartographic-boundary.2020.html
        for details of the shapefiles that will be downloaded on your behalf to
        generate these boundaries.
    remove_water
        If `True` and if with_geometry=True, will query TIGER for AREAWATER shapefiles and
        remove water areas from returned geometry.
    download_contained_within
        A dictionary specifying the geography or geographies that our results
        should be filtered down to be contained within.
    area_threshold
        What fraction of the area of other geographies must be contained
        in our geography to be included. Ignored if `download_contained_within` is
        `None`.
    api_key
        An optional API key. If you don't have or don't use a key, the number
        of calls you can make will be limited to 500 per day.
    variable_cache
        A cache of metadata about variables.
    row_keys
        An optional set of identifier keys to help merge together requests for more than the census API limit of
        50 variables per query. These keys are useful for census datasets such as the Current Population Survey
        where the geographic identifiers do not uniquely identify each row.
    kwargs
        A specification of the geometry that we want data for. For example,
        `state = "*", county = "*"` will download county-level data for
        the entire US.

    Returns
    -------
        A :py:class:`~pd.DataFrame` or `~gpd.GeoDataFrame` containing the requested US Census data.
    """
    if download_contained_within is not None:
        # Put the contained_within context around it.
        return contained_within(
            area_threshold=area_threshold, **download_contained_within
        ).download(
            dataset,
            vintage,
            download_variables,
            group=group,
            leaves_of_group=leaves_of_group,
            set_to_nan=set_to_nan,
            skip_annotations=skip_annotations,
            with_geometry=with_geometry,
            remove_water=remove_water,
            api_key=api_key,
            variable_cache=variable_cache,
            row_keys=row_keys,
            **kwargs,
        )

    if variable_cache is None:
        variable_cache = variables

    # Ensure list operations work
    if row_keys:
        row_keys = list(row_keys)

    # The side effect here is to prime the cache.
    cgeo.geo_path_snake_specs(dataset, vintage)

    if set_to_nan is True:
        set_to_nan = ALL_SPECIAL_VALUES

    # In case they came to us in py format, as kwargs often do.
    kwargs = {
        cgeo.path_component_from_snake(dataset, vintage, k): v
        for k, v in kwargs.items()
    }

    # Parse out the download variables
    download_variables = _parse_download_variables(
        dataset,
        vintage,
        download_variables=download_variables,
        group=group,
        leaves_of_group=leaves_of_group,
        skip_annotations=skip_annotations,
        variable_cache=variable_cache,
    )

    if len(download_variables) <= _MAX_VARIABLES_PER_DOWNLOAD and row_keys:
        warnings.warn(
            "\n The row_keys argument is intended to be used only when the number of requested"
            "\n variables exceeds the Census defined limit of 50"
            "\n The supplied value(s) will be ignored",
            UserWarning,
        )
    # Special case if we are trying to get too many fields.
    if len(download_variables) > _MAX_VARIABLES_PER_DOWNLOAD:
        return _download_multiple(
            dataset,
            vintage,
            download_variables,
            key=api_key,
            census_variables=variable_cache,
            with_geometry=with_geometry,
            row_keys=row_keys,
            **kwargs,
        )

    # Prefetch all the types before we load the data.
    # That way we fail fast if a field is not known.
    _prefetch_variable_types(dataset, vintage, download_variables, variable_cache)
    # Also check that the row_keys, if supplied, are present in the dataset
    if row_keys:
        _prefetch_variable_types(dataset, vintage, row_keys, variable_cache)

    # If we were given a list, join it together into
    # a comma-separated string.
    string_kwargs = {k: _gf2s(v) for k, v in kwargs.items()}

    return _download_remote(
        dataset,
        vintage,
        download_variables=download_variables,
        set_to_nan=set_to_nan,
        with_geometry=with_geometry,
        remove_water=remove_water,
        api_key=api_key,
        variable_cache=variable_cache,
        **string_kwargs,
    )


def _download_remote(
    dataset: str,
    vintage: VintageType,
    *,
    download_variables: List[str],
    set_to_nan: Union[bool, Iterable[float]] = True,
    with_geometry: bool,
    remove_water: bool,
    api_key: Optional[str],
    variable_cache: "VariableCache",
    **kwargs,
) -> Union[pd.DataFrame, gpd.GeoDataFrame]:
    """
    Make the actual remote call to download the data.

    This is the final step after we have parsed out and
    validated the variables and geometry.

    Parameters
    ----------
    dataset
        The dataset to download from. For example `"acs/acs5"`,
        `"dec/pl"`, or `"timeseries/poverty/saipe/schdist"`.
    vintage
        The vintage to download data for. For most data sets this is
        an integer year, for example, `2020`. But for
        a timeseries data set, pass the string `'timeseries'`.
    download_variables
        The census variables to download, for example `["NAME", "B01001_001E"]`.
    set_to_nan
        A list of values that should be set to NaN. Normally these are special
        values that the U.S. Census API sometimes returns. If `True`, then all
        values in :py:ref:`censusdis.values.ALL_SPECIAL_VALUES` will be replaced.
        If `False`, no replacements will be made.
    with_geometry
        If `True` a :py:class:`gpd.GeoDataFrame` will be returned and each row
        will have a geometry that is a cartographic boundary suitable for platting
        a map. See https://www.census.gov/geographies/mapping-files/time-series/geo/cartographic-boundary.2020.html
        for details of the shapefiles that will be downloaded on your behalf to
        generate these boundaries.
    api_key
        An optional API key. If you don't have or don't use a key, the number
        of calls you can make will be limited.
    variable_cache
        A cache of metadata about variables.
    kwargs
        A specification of the geometry that we want data for.

    Returns
    -------
        The downloaded variables, with or without added geometry, as
        either a `pd.DataFrame` or `gpd.GeoDataFrame`.
    """
    url, params, bound_path = census_table_url(
        dataset, vintage, download_variables, api_key=api_key, **kwargs
    )
    df_data = data_from_url(url, params)

    # Coerce the types based on metadata about the variables.
    _coerce_downloaded_variable_types(
        dataset, vintage, download_variables, df_data, variable_cache
    )

    download_variables_upper = [dv.upper() for dv in download_variables]

    # Put the geo fields (STATE, COUNTY, etc...) that came back up front.
    df_data = df_data[
        [col for col in df_data.columns if col not in download_variables_upper]
        + download_variables_upper
    ]

    # NaN out as requested.
    if set_to_nan is True:
        set_to_nan = ALL_SPECIAL_VALUES
    if set_to_nan:
        df_data = df_data.replace(list(set_to_nan), np.nan)

    if with_geometry:
        # We need to get the geometry and merge it in.
        geo_level = bound_path.path_spec.path[-1]
        shapefile_scope = bound_path.bindings[bound_path.path_spec.path[0]]

        gdf_data = add_geography(df_data, vintage, shapefile_scope, geo_level)

        if remove_water:
            gdf_data = clip_water(gdf_data, vintage)

        return gdf_data

    return df_data


def _coerce_downloaded_variable_types(
    dataset: str,
    vintage: VintageType,
    download_variables: List[str],
    df_data: pd.DataFrame,
    variable_cache: "VariableCache",
) -> None:
    """
    Coerce the type of each returned variable (column) in a data frame.

    We look up the type in the metadata in `variable_cache`.

    Parameters
    ----------
    dataset
        The dataset to download from. For example `"acs/acs5"`,
        `"dec/pl"`, or `"timeseries/poverty/saipe/schdist"`.
    vintage
        The vintage to download data for. For most data sets this is
        an integer year, for example, `2020`. But for
        a timeseries data set, pass the string `'timeseries'`.
    download_variables
        The census variables to download, for example `["NAME", "B01001_001E"]`.
    df_data
        The data that came back in JSON form from the census API.
    variable_cache
        A cache of metadata about variables.
    """
    for variable in download_variables:
        # predicateType does not exist in some older data sets like acs/acs3
        # So in that case we just go with what we got in the JSON. But if we
        # have it try to set the type.
        if "predicateType" in variable_cache.get(dataset, vintage, variable):
            field_type = variable_cache.get(dataset, vintage, variable)["predicateType"]

            if field_type == "int" or field_type == "long":
                if df_data[variable].isnull().any():
                    # Some Census data sets put in null in int fields.
                    # We have to go with a float to make this a NaN.
                    # Int has no representation for NaN or None.
                    df_data[variable] = df_data[variable].astype(float, errors="ignore")
                else:
                    try:
                        df_data[variable] = df_data[variable].astype(int)
                    except ValueError:
                        # Sometimes census metadata says int, but they
                        # put in float values anyway, so fall back on
                        # trying to get them as floats.
                        df_data[variable] = df_data[variable].astype(
                            float, errors="ignore"
                        )
                    except OverflowError:
                        # Some long IDs are actually better handled as strings.
                        df_data[variable] = df_data[variable].astype(str)
            elif field_type == "float":
                df_data[variable] = df_data[variable].astype(float)
            elif field_type == "string":
                pass
            else:
                # Leave it as an object?
                pass


def _prefetch_variable_types(
    dataset: str,
    vintage: VintageType,
    download_variables: List[str],
    variable_cache: "VariableCache",
) -> None:
    """
    Prefetch the types of all the variables we are going to try to download.

    This enables us to fail fast and have a better error message about the
    root cause of the issue than if we just blindly put in the variable names
    in the census API request and wait for it to fail.

    Parameters
    ----------
    dataset
        The dataset to download from. For example `"acs/acs5"`,
        `"dec/pl"`, or `"timeseries/poverty/saipe/schdist"`.
    vintage
        The vintage to download data for. For most data sets this is
        an integer year, for example, `2020`. But for
        a timeseries data set, pass the string `'timeseries'`.

    download_variables
        The census variables to download, for example `["NAME", "B01001_001E"]`.
    variable_cache
        A cache of metadata about variables.
    """
    for variable in download_variables:
        try:
            variable_cache.get(dataset, vintage, variable)
        except Exception as exc:
            census_url = CensusApiVariableSource.url(
                dataset, vintage, variable, response_format="html"
            )
            census_variables_url = CensusApiVariableSource.variables_url(
                dataset, vintage, response_format="html"
            )

            raise CensusApiException(
                f"Unable to get metadata on the variable {variable} from the "
                f"dataset {dataset} for year {vintage} from the census API. "
                f"Check the census URL for the variable ({census_url}) to ensure it exists. "
                f"If not found, check {census_variables_url} for all variables in the dataset."
            ) from exc


def _parse_download_variables(
    dataset: str,
    vintage: VintageType,
    *,
    download_variables: Optional[Union[str, Iterable[str]]] = None,
    group: Optional[Union[str, Iterable[str]]] = None,
    leaves_of_group: Optional[Union[str, Iterable[str]]] = None,
    skip_annotations: bool = True,
    variable_cache: Optional["VariableCache"] = None,
) -> List[str]:
    """
    Parse out the full set of download variables.

    These may be encoded in `download_variables`, `group`, and/or `leaves_of_group`.

    Parameters
    ----------
    dataset
        The dataset to download from. For example `"acs/acs5"`,
        `"dec/pl"`, or `"timeseries/poverty/saipe/schdist"`.
    vintage
        The vintage to download data for. For most data sets this is
        an integer year, for example, `2020`. But for
        a timeseries data set, pass the string `'timeseries'`.
    download_variables
        The census variables to download, for example `["NAME", "B01001_001E"]`.
    group
        One or more groups (as defined by the U.S. Census for the data set)
        whose variable values should be downloaded. These are in addition to
        any specified in `download_variables`.
    leaves_of_group
        One or more groups (as defined by the U.S. Census for the data set)
        whose leaf variable values should be downloaded.These are in addition to
        any specified in `download_variables` or `group`. See
        :py:meth:`VariableCache.group_leaves` for more details on the semantics of
        leaves vs. non-leaf group variables.
    skip_annotations
        If `True` try to filter out `group` or `leaves_of_group` variables that are
        annotations rather than actual values. See :py:meth:`VariableCache.group_variables`
        for more details. Variable names passed in `download_variables` are not
        affected by this flag.
    variable_cache
        A cache of metadata about variables.

    Returns
    -------
        The fully expanded list of variables to download.
    """
    # Turn the variables we were given into a list if they are not already.
    if download_variables is None:
        download_variables = []
    elif isinstance(download_variables, str):
        download_variables = [download_variables]
    elif not isinstance(download_variables, list):
        download_variables = list(download_variables)

    if group is None:
        group = []
    elif isinstance(group, str):
        group = [group]

    if leaves_of_group is None:
        leaves_of_group = []
    elif isinstance(leaves_of_group, str):
        leaves_of_group = [leaves_of_group]

    # Add group variables and leaves as appropriate.
    group_variables: List[str] = []
    for group_name in group:
        group_variables = group_variables + variable_cache.group_variables(
            dataset, vintage, group_name, skip_annotations=skip_annotations
        )
    group_leaf_variables: List[str] = []
    for group_name in leaves_of_group:
        group_leaf_variables = group_leaf_variables + variable_cache.group_leaves(
            dataset, vintage, group_name, skip_annotations=skip_annotations
        )

    # Concatenate them all.
    download_variables = download_variables + group_variables + group_leaf_variables

    # Dedup and maintain order.
    download_variables = list(dict.fromkeys(download_variables))

    return download_variables


def census_table_url(
    dataset: str,
    vintage: VintageType,
    download_variables: Iterable[str],
    *,
    api_key: Optional[str] = None,
    **kwargs: cgeo.InSpecType,
) -> Tuple[str, Mapping[str, str], cgeo.BoundGeographyPath]:
    """
    Construct the URL to download data from the U.S. Census API.

    Parameters
    ----------
    dataset
        The dataset to download from. For example `"acs/acs5"`,
        `"dec/pl"`, or `"timeseries/poverty/saipe/schdist"`.
    vintage
        The vintage to download data for. For most data sets this is
        an integer year, for example, `2020`. But for
        a timeseries data set, pass the string `'timeseries'`.
    download_variables
        The census variables to download, for example `["NAME", "B01001_001E"]`.
    api_key
        An optional API key. If you don't have or don't use a key, the number
        of calls you can make will be limited.
    kwargs
        A specification of the geometry that we want data for.

    Returns
    -------
        The URL, parameters and bound path.

    """
    bound_path = _bind_path_if_possible(dataset, vintage, **kwargs)

    query_spec = cgeo.CensusGeographyQuerySpec(
        dataset, vintage, list(download_variables), bound_path, api_key=api_key
    )

    url, params = query_spec.table_url()

    return url, params, bound_path


def _bind_path_if_possible(dataset, vintage, **kwargs):
    """
    Bind the path if possible.

    If not, raise an exception with enough info to fix it.
    """
    bound_path = cgeo.PathSpec.partial_prefix_match(dataset, vintage, **kwargs)
    if bound_path is None:
        raise CensusApiException(
            f"Unable to match the geography specification {kwargs}.\n"
            f"Supported geographies for dataset='{dataset}' in year={vintage} are:\n"
            + "\n".join(
                f"{path_spec}"
                for path_spec in cgeo.geo_path_snake_specs(dataset, vintage).values()
            )
        )
    return bound_path


def geography_names(
    dataset: str,
    vintage: VintageType,
    **kwargs: cgeo.InSpecType,
) -> pd.DataFrame:
    """
    Get the name of a specific geography.

    The arguments are a subset of those to :py:func:`~download`. This
    function is designed to make it easy to fetch the name of a geography
    when we know the FIPS code but want a human-readable name or label for
    display.

    Parameters
    ----------
    dataset
        The dataset to download from. For example `censusdis.datasets.ACS5`.
    vintage
        The vintage to download data for. For example, `2020`.
    kwargs
        A specification of the geometry that we want data for. For example,
        `state = "34", county = "017"` will download the name of Hudson County,
        New Jersey.

    Returns
    -------
        A dataframe with columns specifying the geography and one for the name.
        All column names will be in ALL CAPS.
    """
    df = download(dataset, vintage, ["NAME"], **kwargs)

    return df


def geographies(dataset: str, vintage: VintageType) -> List[List[str]]:
    """
    Determine what geographies are supported for a dataset and vintage.

    This utility gives us a list of the different geography
    keywords we can use in calls to :py:func:`download` with
    for the given dataset and vintage.

    Parameters
    ----------
    dataset
        The dataset to download from. For example `"acs/acs5"`,
        `"dec/pl"`, or `"timeseries/poverty/saipe/schdist"`.
    vintage
        The vintage to download data for. For most data sets this is
        an integer year, for example, `2020`. But for
        a timeseries data set, pass the string `'timeseries'`.

    Returns
    -------
        A list of lists of geography keywords. Each element
        of the outer list is a list of keywords that can be
        used together.
    """
    return list(cgeo.geo_path_snake_specs(dataset, vintage).values())


variables = VariableCache()


def _intersecting_geos_kws(
    dataset: str,
    vintage: VintageType,
    containing_geo_kwargs: cgeo.InSpecType,
    **kwargs: cgeo.InSpecType,
) -> cgeo.InSpecType:
    """
    Construct geography keywords for intersecting geographies.

    Parameters
    ----------
    dataset
        The dataset to download from. For example `"acs/acs5"`,
        `"dec/pl"`, or `"timeseries/poverty/saipe/schdist"`. There are
        symbolic names for datasets, like `ACS5` for `"acs/acs5"
        in :py:module:`censusdis.datasets`.
    vintage
        The vintage to download data for. For most data sets this is
        an integer year, for example, `2020`. But for
        a timeseries data set, pass the string `'timeseries'`.
    containing_geo_kwargs
        Geographic keywords specifying the containing geography that we are
        looking for intersections with. For example
        `dict(metropolitan_statistical_area_micropolitan_statistical_area="35620")`
        for the New York area CBSA.
    kwargs
        A specification of the geometry that we want data for, limited to those
        geographies that are contained in the geography specified by `containing_geo_kwargs`.
        For example, `state="*", county="*", tract="*"` will specifies county-level data for
        all counties contained in the containing geography.

    Returns
    -------
        A dictionary of geographic keywords suitable for passing to :py:func:`~download`.
    """
    # This is a fast short circuit if there is only one
    # element of kwargs or the first component
    # is already specified. The former is since we will have to
    # query with the kwargs as they are, so we might as well
    # just let our caller do it. The second case is because
    # we might trim down the list, but more likely the user
    # double specified at the top level, like state=.
    if len(kwargs) == 1 or list(kwargs.values())[0] != "*":
        return kwargs

    # Download the geometry of the outer scope.
    gdf_within = download(
        dataset, vintage, ["NAME"], with_geometry=True, **containing_geo_kwargs
    )

    # See if we can find a matching path spec.
    bound_path = _bind_path_if_possible(dataset, vintage, **kwargs)

    # Get the geography for the outermost level of the match.
    first_binding = list(bound_path.bindings.items())[0]

    containing_geo_kwargs = {first_binding[0]: first_binding[1]}

    gdf_first_binding = download(
        dataset, vintage, ["NAME"], with_geometry=True, **containing_geo_kwargs
    )

    # Which of the first binding geographies intersect
    # the area we want our final geographies to be in.
    gdf_intersects = gdf_first_binding.sjoin(
        gdf_within, lsuffix="FIRST", rsuffix="within"
    )

    col_name = first_binding[0].replace(" ", "_").upper()
    if col_name not in gdf_intersects.columns:
        col_name = f"{col_name}_FIRST"

    intersecting_geographies = list(gdf_intersects[col_name].unique())

    # Short circuit if there are a massive number of intersection
    # geps. In this case, we'll just leave things as they came with
    # the leading '*' and query them all. Otherwise the URL gets super
    # long and things go a little crazy. This can happen with zip code
    # tabulation areas.
    if len(intersecting_geographies) > 20:
        return dict(**kwargs)

    intersecting_geographies = [
        geo[:-6] if geo.endswith("_FIRST") else geo for geo in intersecting_geographies
    ]

    geo = dict(bound_path.bindings)

    geo[first_binding[0]] = intersecting_geographies

    return geo


class ContainedWithin:
    """A representation of a geography that we want to query some other geographies that are contained within."""

    def __init__(self, area_threshold: float = 0.8, **kwargs: cgeo.InSpecType):
        """
        Construct a representation of a geography that we want to query some other geographies contained within.

        Parameters
        ----------
        area_threshold
            What fraction of the area of other geographies must be contained
            in our geography to be included.
        kwargs
            A specification of the geometry that we want data for geometries
            that are contained within. For example,
            `state = "NJ", place = "01960"` will specify the city of Asbury Park, NJ.
        """
        self._area_threshold = area_threshold
        self._containing_kwargs = kwargs

    def __eq__(self, other) -> bool:
        """Are two objects equal."""
        if not isinstance(other, ContainedWithin):
            return False

        return (
            self._area_threshold == other._area_threshold
            and self._containing_kwargs == other._containing_kwargs
        )

    def __enter__(self) -> "ContainedWithin":
        """
        Enter the context.

        Returns
        -------
            The ContainedWithin object for use within the context.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Exit the context."""
        pass

    def download(
        self,
        dataset: str,
        vintage: VintageType,
        download_variables: Optional[Union[str, Iterable[str]]] = None,
        *,
        group: Optional[Union[str, Iterable[str]]] = None,
        leaves_of_group: Optional[Union[str, Iterable[str]]] = None,
        set_to_nan: Union[bool, Iterable[int]] = True,
        skip_annotations: bool = True,
        with_geometry: bool = False,
        remove_water: bool = False,
        api_key: Optional[str] = None,
        variable_cache: Optional["VariableCache"] = None,
        row_keys: Optional[Union[str, Iterable[str]]] = None,
        **kwargs: cgeo.InSpecType,
    ) -> Union[pd.DataFrame, gpd.GeoDataFrame]:
        """
        Download data for geographies contained within a containing geography.

        Parameters
        ----------
        dataset
            The dataset to download from. For example `"acs/acs5"`,
            `"dec/pl"`, or `"timeseries/poverty/saipe/schdist"`. There are
            symbolic names for datasets, like `ACS5` for `"acs/acs5"
            in :py:module:`censusdis.datasets`.
        vintage
            The vintage to download data for. For most data sets this is
            an integer year, for example, `2020`. But for
            a timeseries data set, pass the string `'timeseries'`.
        download_variables
            The census variables to download, for example `["NAME", "B01001_001E"]`.
        group
            One or more groups (as defined by the U.S. Census for the data set)
            whose variable values should be downloaded. These are in addition to
            any specified in `download_variables`.
        leaves_of_group
            One or more groups (as defined by the U.S. Census for the data set)
            whose leaf variable values should be downloaded.These are in addition to
            any specified in `download_variables` or `group`. See
            :py:meth:`VariableCache.group_leaves` for more details on the semantics of
            leaves vs. non-leaf group variables.
        set_to_nan
            A list of values that should be set to NaN. Normally these are special
            values that the U.S. Census API sometimes returns. If `True`, then all
            values in :py:ref:`censusdis.values.ALL_SPECIAL_VALUES` will be replaced.
            If `False`, no replacements will be made.
        skip_annotations
            If `True` try to filter out `group` or `leaves_of_group` variables that are
            annotations rather than actual values. See :py:meth:`VariableCache.group_variables`
            for more details. Variable names passed in `download_variables` are not
            affected by this flag.
        with_geometry
            If `True` a :py:class:`gpd.GeoDataFrame` will be returned and each row
            will have a geometry that is a cartographic boundary suitable for platting
            a map. See https://www.census.gov/geographies/mapping-files/time-series/geo/cartographic-boundary.2020.html
            for details of the shapefiles that will be downloaded on your behalf to
            generate these boundaries.
        remove_water
            If `True` and if with_geometry=True, will query TIGER for AREAWATER shapefiles and
            remove water areas from returned geometry.
        api_key
            An optional API key. If you don't have or don't use a key, the number
            of calls you can make will be limited to 500 per day.
        variable_cache
            A cache of metadata about variables.
        row_keys
            An optional set of identifier keys to help merge together requests for more than the census API limit of
            50 variables per query. These keys are useful for census datasets such as the Current Population Survey
            where the geographic identifiers do not uniquely identify each row.
        kwargs
            A specification of the geometry that we want data for. For example,
            `state = "*", county = "*"` will download county-level data for
            the entire US.

        Returns
        -------
            A :py:class:`~pd.DataFrame` or `~gpd.GeoDataFrame` containing the requested US Census data.
        """
        geos_kwargs = _intersecting_geos_kws(
            dataset, vintage, self._containing_kwargs, **kwargs
        )

        gdf = download(
            dataset,
            vintage,
            download_variables,
            group=group,
            leaves_of_group=leaves_of_group,
            set_to_nan=set_to_nan,
            skip_annotations=skip_annotations,
            with_geometry=True,
            remove_water=remove_water,
            api_key=api_key,
            variable_cache=variable_cache,
            row_keys=row_keys,
            **geos_kwargs,
        )

        # See which of these geometries are mostly contained by
        # the geography we want to be within.

        gdf_container = download(
            dataset, vintage, ["NAME"], with_geometry=True, **self._containing_kwargs
        ).drop("NAME", axis="columns")

        gdf_contained = cmap.sjoin_mostly_contains(
            gdf_container, gdf, area_threshold=self._area_threshold
        )

        # Drop all the large container columns we don't need.
        gdf_contained = gdf_contained[
            [col for col in gdf_contained.columns if not col.endswith("_large")]
        ].reset_index(drop=True)

        # Drop the "_small" suffix.

        gdf_contained.rename(
            lambda col: col[:-6] if col.endswith("_small") else col,
            axis="columns",
            inplace=True,
        )

        if with_geometry:
            # Keep the columns from the larger result.
            return gdf_contained[
                [
                    col
                    for col in gdf_container.columns
                    if col in gdf_contained.columns and col != "geometry"
                ]
                + [
                    col
                    for col in gdf_contained.columns
                    if col not in gdf_container.columns or col == "geometry"
                ]
            ]
        else:
            # Drop the geometry and return a `pd.DataFrame`
            return pd.DataFrame(
                gdf_contained[
                    [
                        col
                        for col in gdf_container.columns
                        if col in gdf_contained.columns and col != "geometry"
                    ]
                    + [
                        col
                        for col in gdf_contained.columns
                        if col not in gdf_container.columns and col != "geometry"
                    ]
                ]
            )


def contained_within(
    area_threshold: float = 0.8, **kwargs: cgeo.InSpecType
) -> ContainedWithin:
    """
    Construct a representation of a geography that we want to query some other geographies contained within.

    Parameters
    ----------
    area_threshold
        What fraction of the area of other geographies must be contained
        in our geography to be included.
    kwargs
        A specification of the geometry that we want data for geometries
        that are contained within. For example,
        `state = "NJ", place = "01960"` will specify the city of Asbury Park, NJ.
    """
    return ContainedWithin(area_threshold=area_threshold, **kwargs)


def add_inferred_geography(
    df_data: pd.DataFrame, year: Optional[int] = None
) -> gpd.GeoDataFrame:
    """
    Infer the geography level of the given dataframe.

    Add geometry to each row for the inferred level.

    See Also
    --------
        :py:ref:`~infer_geo_level` for more on how inference is done.

    Parameters
    ----------
    df_data
        A dataframe of variables with one or more columns that
        can be used to infer what geometry level the rows represent.
    year
        The year for which to fetch geometries. We need this
        because they change over time. If `None`, look for a
        `'YEAR'` column in `df_data` and possibly add different
        geometries for different years as needed.

    Returns
    -------
        A geo data frame containing the original data augmented with
        the appropriate geometry for each row.
    """
    if year is None:
        # We'll try to get the year out of the data.
        if "YEAR" not in df_data.columns:
            raise ValueError(
                "If year is None then there must be a `YEAR` column in the data."
            )

        return gpd.GeoDataFrame(
            df_data.groupby("YEAR", group_keys=False)
            .apply(lambda df_group: add_inferred_geography(df_group, df_group.name))
            .reset_index(drop=True)
        )

    geo_level = infer_geo_level(year, df_data)

    (
        shapefile_scope,
        _,
        shapefile_scope_columns,
        _,
    ) = geo_query_from_data_query_inner_geo(year, geo_level)

    if shapefile_scope is not None:
        # The scope is the same across the board.
        gdf = add_geography(df_data, year, shapefile_scope, geo_level)
        return gdf

    # We have to group by different values of the shapefile
    # scope from the appropriate column and add the right
    # geography to each group.
    shapefile_scope_column = shapefile_scope_columns[0]

    df_with_geo = (
        df_data.groupby(shapefile_scope_column, group_keys=False)
        .apply(lambda g: add_geography(g, year, g.name, geo_level))
        .reset_index(drop=True)
    )

    gdf = gpd.GeoDataFrame(df_with_geo)

    return gdf


certificates = censusdis.impl.fetch.certificates
